using System;
using System.IO;
using System.Text;
using System.Text.RegularExpressions;

using JustinIO;

namespace Bootloader
{
	/// <summary>
	/// Encapsulates an interface to read and write firmware .HEX files
	/// </summary>
	class Firmware
	{
		private string[] pages;
		private string version = "";
		private string model = "";

		public Firmware ()
		{
			pages = new string[AYUCR.MAXADDR + 1];
		}

		public string Version
		{
			get 
			{
				return version;
			}
		}

		public string Model
		{
			get
			{
				return model;
			}
		}


		/// <summary>
		/// Computes a sum on a string of hex digits
		/// </summary>
		/// <param name="data"></param>
		/// <returns>int</returns>
		public int HexToSum (string data)
		{
			int sum = 0;

			for (int i = 0; i <= data.Length / 2 - 1; i++) 
			{
				sum += Convert.ToByte ("0x" + data.Substring (i * 2, 2), 16);
			}

			return sum;
		}

		public void SetPage (ushort page, string data)
		{
			pages[page] = data;
		}

		public string GetPage (ushort page)
		{
			return pages[page];
		}

		public void ClearPage (ushort page)
		{
			pages[page] = null;
		}

		/// <summary>
		/// Reads a firmware image file
		/// </summary>
		/// <param name="filename">File to read</param>
		public void Read (string filename)
		{
			// Create an instance of StreamReader to read from the .HEX file.
			// The using statement also closes the StreamReader.
			using (StreamReader sr = new StreamReader(filename)) 
			{
				Regex rx = new Regex(@"^:(\S\S)(\S\S)(\S\S)(\S\S)(\S*)(\S\S)", RegexOptions.IgnoreCase);

				String line;

				// Read and display lines from the file until the end of 
				// the file is reached.
				while ((line = sr.ReadLine()) != null) 
				{
					Match m = rx.Match(line);
					GroupCollection g = m.Groups;

					string scount = "0x" + g[1];
					byte count = Convert.ToByte ("0x" + g[1], 16);
					byte adrhi = Convert.ToByte ("0x" + g[2], 16);
					byte adrlo = Convert.ToByte ("0x" + g[3], 16);
					byte type = Convert.ToByte ("0x" + g[4], 16);
					string data = g[5].ToString ();
					byte checksum = Convert.ToByte ("0x" + g[6], 16);

					int address = (adrhi << 8) + adrlo;
					int page = (address + 1) / AYUCR.PAGESIZE;
					int offset = address % AYUCR.PAGESIZE;

					// Calculate checksum based on data read
					int sum = (count + adrlo + adrhi + type + HexToSum (data));
					if (((~sum + 1) & 0xFF) != checksum)
						throw (new ApplicationException("Bad checksum in file"));

					sum = (~sum + 1) & 0xFF;

					// Add the data to the appropriate page.  Ignore non-data records and
					// addresses out of range.
					if (type == 0 && page < pages.Length) 
					{
						// Allocate a page if null
						if (pages[page] == null)
							pages[page] = "".PadRight (AYUCR.PAGESIZE * 2);

						pages[page] = pages[page].Substring(0, offset*2) + data + pages[page].Substring(offset*2 + data.Length);
					}
				}
			}

			UpdateModelVersion ();
		}

		/// <summary>
		/// Write the firmware image to a file
		/// </summary>
		/// <param name="filename">File to write to</param>
		public void Write (string filename)
		{
			// Create a regex to break the page string into blocks
			Regex rx = new Regex(@"(\s*)(\S+)(.*)", RegexOptions.IgnoreCase);
			
			using (StreamWriter hex = new StreamWriter (filename)) 
			{
				// Loop over all pages
				for (int page = 0; page <= AYUCR.MAXADDR; page++) 
				{
					if (pages[page] != null) 
					{
						// break each page into blocks of 16 bytes
						for (int block = 0; block <= 3; block++) 
						{
							int offset = block * AYUCR.PAGESIZE/2;
							int addr = page * AYUCR.PAGESIZE + offset/2;
							string page_string = pages[page].Substring (offset, AYUCR.PAGESIZE/2);

							// Loop over block until everything has been dumped
							for (;;) 
							{
								Match m = rx.Match(page_string);
								GroupCollection g = m.Groups;

								string pre = g[1].ToString ();
								string data = g[2].ToString ();
								string rest = g[3].ToString ();

								if (data == "")
									break;

								// Update address to account for skipped bytes
								addr += pre.Length/2;

								int sum = data.Length/2 + (addr >> 8) + (addr & 0xFF) + HexToSum (data);
								int checksum = (~sum + 1) & 0xFF;

								hex.WriteLine (":{0:X2}{1:X4}{2:X2}{3}{4:X2}", data.Length/2, addr, 0x00, data, checksum);

								addr += data.Length/2;

								page_string = rest;
							}
						}
					}
				}

				hex.WriteLine (":00000001FF");
			}
		}
	
		/// <summary>
		/// Dump the contents of the page buffer to a string
		/// </summary>
		public string Dump ()
		{
			string output = "";

			for (int page = 0; page < pages.Length; page++) 
			{
				if (pages[page] != null) 
				{
					for (int block = 0; block <= 7; block++) 
					{
						int offset = block * AYUCR.PAGESIZE/4;
						int addr = page * AYUCR.PAGESIZE + offset/2;
						string page_string = pages[page].Substring (offset, AYUCR.PAGESIZE/4);

						output += String.Format ("{0:X4} ", addr);

						for (int i = 0; i < 8; i++) 
						{
							output += pages[page].Substring (offset + i * 2, 2) + " ";
						}
						output += "  ";

						// Text representation
						for (int i = 0; i < 8; i++) 
						{
							string val = pages[page].Substring (offset + i * 2, 2);
							char c = val != "  " ? Convert.ToChar (Convert.ToByte ("0x" + val, 16)) : '\0';
							
							output += Char.IsControl (c) ? '.' : c;
						}
						
						output += "\n";
					}
				}
			}

			return output;
		}

		/// <summary>
		/// Compares two firmware objects within a specified range
		/// </summary>
		/// <param name="other">The firmware object to compare against</param>
		/// <param name="low">lowest page in range</param>
		/// <param name="high">highest page in range</param>
		/// <returns>true if images are the same</returns>
		public bool Compare (Firmware other, int low, int high)
		{
			for (int i = low; i <= high; i++) 
			{
				if (pages[i] != other.pages[i])
					return false;
			}

			return true;
		}

		/// <summary>
		/// Extracts the model and version info from the firmware image
		/// </summary>
		private void UpdateModelVersion ()
		{
			// 563432342E34303463340034
			// Regular expression to look for a series of RETLW instructions 
			// starting with RETLW 'V' and ending with RETLW '\0'
			Regex rx = new Regex(@"(5634\S+?)0034.*?(\S\S34\S+?)0034", RegexOptions.IgnoreCase);
		
			string first_page = pages[1];

			if (first_page != null) 
			{
				// Look for version string
				Match m = rx.Match(first_page);
				GroupCollection g = m.Groups;
				string segment = g[1].ToString ();

				// Extract the ascii string from the segment of instructions
				version = "";
				for (int i = 0; i < segment.Length; i += 4) 
				{
					version += Convert.ToChar (Convert.ToByte ("0x" + segment.Substring (i, 2), 16));
				}

				segment = g[2].ToString ();

				// Extract the ascii string from the segment of instructions
				model = "";
				for (int i = 0; i < segment.Length; i += 4) 
				{
					model += Convert.ToChar (Convert.ToByte ("0x" + segment.Substring (i, 2), 16));
				}
			}
		}
	}

	class AYUCR : IDisposable	
	{
		public static ushort PAGESIZE = 64;
		public static ushort MAXADDR = 0x3F;
		public static string NULLVAL = "FF";
		public static ushort MINUSER = 0x01;
		public static ushort MAXUSER = 0x37;
		public static ushort MINBOOT = 0x38;
		public static ushort MAXBOOT = 0x3F;

		public string model;
		public string version;
		private CommPort port;
 
		/// <summary>
		/// Read and test for the OK prompt
		/// </summary>
		/// <returns>true if OK found</returns>
		private bool GetOK ()
		{
			byte[] ok = port.Read (2);

			return ok.Length == 2 && ok[0] == 'O' && ok[1] == 'K';
		}

		/// <summary>
		/// Connects to controller, captures model/version info and starts bootloader
		/// </summary>
		/// <returns>true on successs</returns>
		public void Connect (int port_num)
		{
			if (port != null && port.Opened) 
			{
				throw(new ApplicationException("Comm Port Already Open"));
			}

			// Initialize the comm settings
			port = new CommPort ();

			port.PortNum = port_num;
			port.BaudRate = 2400;
			port.ReadTimeout = 1000; // 1 sec

			port.Open ();

			// Model String
			port.Write ("/M");
			port.ReadLine ();
			model = port.ReadLine ();
			GetOK ();

			// Version String
			port.Write ("/V");
			port.ReadLine ();
			version = port.ReadLine ();
			GetOK ();

			// Make sure bootloader is available
			if (version[0] == 'V' && version[1] == '1') 
			{
				throw (new ApplicationException ("Bootloader unavailable in firmware versions prior to 2.0"));
			}

			// Start Bootloader
			for (int i = 0; i < 3; i++) 
			{
				port.Write ("/B");
				if (port.Read () == "K")
					return;
			}

			// port.Close ();
			throw (new ApplicationException("Cannot find bootloader"));
		}

		/// <summary>
		/// Disconnect from controller
		/// </summary>
		public void Dispose ()
		{
			port.Close ();
		}

		/// <summary>
		/// Form the address bytes LO/HI
		/// </summary>
		/// <param name="page">Page number</param>
		/// <returns></returns>
		private byte[] Address (ushort page)
		{
			ushort addr = (ushort) (page * PAGESIZE / 2);

			byte[] address = { (byte) (addr & 0xFF), (byte) (addr / 0x100) };

			return address;

			// Address is lowbyte, highbyte
		//	return ((char) (addr & 0xFF)).ToString () + ((char) (addr / 0xFF)).ToString ();
		}

		private byte Checksum (byte[] data)
		{
			byte checksum = 0;

			//char[] chars = data.ToCharArray ();
			for (int i = 0; i < data.Length; i++)
				unchecked { checksum += data[i]; }

			return checksum;
		}

		/// <summary>
		/// Form a command string
		/// </summary>
		/// <param name="command">single character command</param>
		/// <param name="page">page number reference</param>
		/// <returns>string containing the command, arguments and checksum byte</returns>
		private byte[] Command (char command, ushort page)
		{
			byte[] address = Address (page);
			byte[] cmd = new byte[4];

			cmd[0] = (byte) command;
			cmd[1] = address[0];
			cmd[2] = address[1];
			cmd[3] = Checksum (address);

			return cmd;
		}

		private string Wait_K ()
		{
            string data = "", prompt;
			int timeout = 5;

			for (;;) 
			{
				prompt = port.Read ();
				data += prompt;
				if (prompt == "K")
					break;

				// Check timeout
				if (prompt == "" && timeout-- == 0)
					break;
			}

			return data;
		}

		/// <summary>
		/// Read a page via the bootloader
		/// </summary>
		/// <param name="page">page number</param>
		/// <returns>string (128 chars) of 64 byte pairs</returns>
		public byte[] Read (ushort page)
		{
			port.Write (Command ('R', page));

			return port.Read (64);
		}

		/// <summary>
		/// Erase a page
		/// </summary>
		/// <param name="page">page to erase</param>
		public void Erase (ushort page)
		{
			port.Write (Command ('E', page));
		}

		/// <summary>
		/// Writes a single page to bootloader
		/// </summary>
		/// <param name="page">page number</param>
		/// <param name="data">page as a 64 character string</param>
		public void Write (ushort page, byte[] data)
		{
			if (data.Length != PAGESIZE) 
				throw (new ApplicationException("Invalid page size"));

			byte[] address = Address (page);

			// Form the complete command W(1) + A(2) + D(64) + C1
			byte[] cmd = new byte[68];
			cmd[0] = (byte) 'W';
			cmd[1] = address[0];
			cmd[2] = address[1];

			for (int i = 0; i < data.Length; i++)
				cmd[i + 3] = data[i];

			// Compute checksum.
			unchecked { cmd[67] = (byte) (Checksum (address) + Checksum (data)); }
			
			port.Write (cmd);
		}

		/// <summary>
		/// Write firmware image to bootloader
		/// </summary>
		/// <param name="firmware">firmware to write</param>
		public void Write (Firmware firmware, System.Windows.Forms.ProgressBar progress)
		{
			// Write all non protected pages
			for (ushort page = MINUSER; page <= MAXUSER; page++) {
				string page_string = firmware.GetPage (page);

				if (page_string == null) 
					Erase (page);
				else 
				{
					page_string = page_string.Replace ("    ", "FF3F");
					
					Write (page, HexCon.StringToByte (page_string));
				}

				string prompt = Wait_K ();
				if (prompt != "K")
					throw(new ApplicationException("Bootloader Error: " + prompt));

				if (progress != null)
					progress.PerformStep ();
			}
		}

		/// <summary>
		/// Read firmware image from bootloader
		/// </summary>
		/// <returns>Firmware image</returns>
		public Firmware Read (System.Windows.Forms.ProgressBar progress)
		{
			Firmware firmware = new Firmware ();

			// Read all pages and create a list
			for (ushort page = 0; page <= MAXADDR; page++) 
			{
				byte[] data = Read (page);

				// Check for an error
				if (data[0] == 'C' && data[1] == 'K')
					throw (new ApplicationException("Bootloader Error: page " + page));

				byte[] checksum = port.Read (1);

				Wait_K ();

				// Check checksum
				if (checksum[0] != Checksum (data))
					throw (new ApplicationException("Bootloader Error: " + data));

				// Convert bytes to string representation
				string page_string = "";
				foreach (byte b in data)
					page_string += String.Format("{0:X2}", b);

				// Replace default bytes with blanks
				page_string = page_string.Replace ("FF3F", "    ");

				// Store the new page as long as it isn't completely empty
				if (page_string.Trim () == "")
					firmware.ClearPage (page);
				else
					firmware.SetPage (page, page_string);

				if (progress != null)
					progress.PerformStep ();
			}

			return firmware;
		}
	}
}